
/*****************************************************************************/
/*                                                                           */
/*  XARXES DE COMPUTADORS I                                                  */
/*                                                                           */
/*  Practica 3: Image Hunter                                                 */
/*  Extraccion de todas las imagenes de la pagina de inicio del servidor     */
/*  especificado, y almacenamiento en disco de cada una de ellas.            */
/*                                                                           */
/*  Fichero cola.c: Modulo que incluye las todas las operaciones para poder  */  
/*                  manejar las estructura de datos Cola, siendo sta una    */
/*                  cola dinmica, encadenada y con punto de inters (PDI).  */
/*                                                                           */
/*  Realizado por: Francisco Jose Aguilar Celdran y Sergio Blanco Cuaresma   */
/*                                                                           */
/*****************************************************************************/

#define MODULO_COLA
#include <stdlib.h>
#include <stdio.h>
#include "cola.h"
    
/* --------------------------------- */
/* CREAR_COLA: Crea una nueva cola.  */
/* --------------------------------- */
         
Cola *crear_cola(void)
{
   Cola *nueva;

   nueva =(Cola *)malloc(sizeof(Cola));
   nueva -> primero = NULL;
   nueva -> ultimo = NULL;
   nueva -> pdi = NULL;
   nueva -> total = 0;
   return nueva;
}

/* -------------------------------------------------- */
/* COLA_VACIA: responde si la cola esta o no vacia.   */
/* -------------------------------------------------- */

unsigned int cola_vacia(Cola *c)
{
   if(c -> primero == NULL) 
      return 1;
   else 
      return 0;
}

/* ----------------------------------------------------------------- */
/* ENCOLAR: aade como ultimo elemento de la cola el nodo recibido.  */
/* ----------------------------------------------------------------- */

void encolar(Cola *c,char *valores)
{
   Nodo *nuevo_nodo;
   
   nuevo_nodo = (Nodo*)malloc(sizeof(Nodo));
   nuevo_nodo -> nodo = valores;
   nuevo_nodo -> anterior = NULL;
   if (cola_vacia(c))
   {
      c -> primero = nuevo_nodo;
      c -> ultimo = nuevo_nodo;
      c -> pdi = nuevo_nodo;
   }
   else
   {
      c -> ultimo -> anterior = nuevo_nodo;
      c -> ultimo = nuevo_nodo;
   }
   c -> total++;
}

/* --------------------------------------------------- */
/* DESENCOLAR: elimina el primer elemento de la cola.  */
/* --------------------------------------------------- */

void desencolar(Cola *c)
{
   Nodo *aux;
   
   if (!cola_vacia(c))
   {  
      if (c -> primero -> anterior == NULL)
      {
         c -> primero = NULL; 
         c -> ultimo = NULL;
         c -> pdi = NULL;
      }  
      else
      {
         aux = c -> primero;
         c -> primero = c -> primero -> anterior;
         free(aux);
      }
   }
   c -> total--;
}

/* ------------------------------------------------------- */
/* PRIMERO: retorna el primer elemento aadido a la cola.  */
/* ------------------------------------------------------- */

unsigned char *primero(Cola *c)
{
   if (c -> primero != NULL)
      return c -> primero -> nodo;
   else
      return NULL;
}

/* ------------------------------------------------------------------------------- */
/* REINICIAR PDI: establece el Punto de Interes en en primer elemento de la cola.  */
/* ------------------------------------------------------------------------------- */

void reiniciar_PDI(Cola *c)
{
   c -> pdi = c -> primero;   
}

/* ----------------------------------------------------------------------------------- */
/* OBTENER ELEMENTO PDI: retorna el nodo apuntado por el Punto de Interes de la cola.  */
/* ----------------------------------------------------------------------------------- */

unsigned char *obtener_elemento_PDI(Cola *c)
{
   if (c -> pdi != NULL)
      return c -> pdi -> nodo;
   else
      return NULL;
}

/* -------------------------------------------------------------------------- */
/* AVANZAR PDI: apunta el Punto de Interes al elemento siguiente de la Cola.  */
/* -------------------------------------------------------------------------- */

int avanzar_PDI(Cola *c)
{
   if (c -> pdi != NULL)
   {
      c -> pdi = c -> pdi -> anterior; 
      return 1;
   }
   else
      return 0;   
}

/* --------------------------------------------------------------------------- */
/* VACIAR COLA: elimina todos los elementos de la Cola, liberando su memoria.  */
/* --------------------------------------------------------------------------- */

void vaciar_cola(Cola *c)
{
   while (!cola_vacia(c)) {
      desencolar(c);
   }
}



   
