
/*****************************************************************************/
/*                                                                           */
/*  XARXES DE COMPUTADORS I                                                  */
/*                                                                           */
/*  Practica 3: Image Hunter (Cazador de Imagenes)                           */
/*  Extraccion de todas las imagenes de la pagina de inicio del servidor     */
/*  especificado, y almacenamiento en disco de cada una de ellas.            */
/*                                                                           */
/*  Fichero image_hunter.c: incluye el las funciones que implementan         */
/*                          la logica de la aplicacion.                      */
/*                                                                           */
/*  Realizado por: Francisco Jose Aguilar Celdran y Sergio Blanco Cuaresma   */
/*                                                                           */
/*****************************************************************************/


#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/dir.h>
#include <netdb.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include "cola.h"
#include "image_hunter.h"
#include "mensajes.h"

/* --------------------------------------------------------------------------------- */
/* DISPERSION_IMAGEN: funcion de hashing que aplica dispersion del nombre de imagen  */
/*                    recibido mediante suma de los codigos ASCII de la cadena de    */
/*                    caracteres, con el fin de acceder a la posicion de la tabla    */
/*                    de nombres de imagenes (tabla de hash) con coste constante.    */
/* --------------------------------------------------------------------------------- */

int dispersion_imagen (char *nombre_imagen)
{
   int longitud = strlen(nombre_imagen);
   int i, valor = 0;
   
   for(i = 0; i < longitud; i++) {
      valor += (int)nombre_imagen[i];
   }
   valor = valor % TAM_TABLA_HASH;
   return valor;
}

/* --------------------------------------------------------------------------------------- */
/* DETERMINAR_IMAGEN_EXISTENTE: confirma con coste constante si la imagen recibida ya ha   */
/*                              sido tratada previamente. Se utiliza para ello una tabla   */ 
/*                              de hash, donde cada posicion de la tabla es una cola, que  */
/*                              permite gestionar las posibles colisiones de nombres para  */
/*                              una misma posicion de la tabla de hashing.                 */
/* --------------------------------------------------------------------------------------- */

int determinar_imagen_existente (Cola **tabla_hash, int posicion_tabla, char *nombre_imagen)
{
   int imagen_repetida = 0;
   char *posible_colision;

   if (tabla_hash[posicion_tabla] == NULL) {
      tabla_hash[posicion_tabla] = (Cola *)crear_cola();
      encolar(tabla_hash[posicion_tabla],nombre_imagen);
   }
   else {
      reiniciar_PDI(tabla_hash[posicion_tabla]); 
      posible_colision = obtener_elemento_PDI(tabla_hash[posicion_tabla]);
      while ((posible_colision != NULL) && (!imagen_repetida)) {
         if (strlen(posible_colision) == strlen(nombre_imagen)) {
            if (strcasecmp(posible_colision,nombre_imagen) == 0) {
               imagen_repetida = 1;
            }
         }
         avanzar_PDI(tabla_hash[posicion_tabla]);
         posible_colision = obtener_elemento_PDI(tabla_hash[posicion_tabla]);
      }
      if (!imagen_repetida) {
         encolar(tabla_hash[posicion_tabla],nombre_imagen);
      }
   }
   return imagen_repetida;
}

/* ------------------------------------------------------------------------------------- */
/* COMPARAR_ETIQUETAS: determina si los 'num_cars' primeros caracteres de la cadena      */
/*                     de texto 'html', a partir del 'indice' especificado, son iguales  */
/*                     a la etiqueta 'tag1', o bien a la etiqueta 'tag2'.                */
/* ------------------------------------------------------------------------------------- */

int comparar_etiquetas (char *html, char *tag1, char *tag2, int num_cars, int indice)
{
   int contador = 0;
   int iguales = 1;
   
   while ((contador < num_cars) && (iguales == 1)) {
      if (!(html[indice + contador] == tag1[contador]) && !(html[indice + contador] == tag2[contador])) {
         iguales = 0;
      }
      else {
         contador++;
      }
   }
   return iguales;
}

/* ----------------------------------------------------------------------------------- */
/* BUSCADOR_IMAGENES: funcion que extrae las ubicaciones de todas las imagenes         */
/*                    contenidas en el texto html recibido. Dichas ubicaciones         */
/*                    se van encolando en una cola que es devuelta como resultado.     */
/*                    Busca las etiquetas <IMG> y dentro de estas el atributo <SRC>.   */
/*                    No tiene en cuenta el texto que hay dentro de un <SCRIPT>.       */
/*                    Se omiten imagenes repetidas, para ello se utiliza una tabla     */
/*                    de hashing, consultando en coste constante si existe la imagen.  */
/* ----------------------------------------------------------------------------------- */

Cola *buscador_imagenes (char *texto_html)
{
   int j;
   long i = 0;
   int contador = 1;
   long longitud = 0;
   int script_activo = 0;
   Cola *cola;
   int fin_etiqueta;
   char caracter_fin_imagen;
   char caracter_auxiliar;
   unsigned char *imagen;
   int valor_dispersion;
   int imagen_repetida;
   Cola *tabla_hash[TAM_TABLA_HASH];
   
   for (j = 0; j < TAM_TABLA_HASH; j++) {
      tabla_hash[j] = NULL;
   }
   longitud = strlen(texto_html);
   cola = (Cola *)crear_cola();
   while (i < longitud - 8) {
      if (texto_html[i] == '<') {
         i++;
         if (texto_html[i] == '/') {
            i++;
            if (comparar_etiquetas(texto_html,"script","SCRIPT",6,i) ||     // Fin de script, se siguen buscando
                comparar_etiquetas(texto_html,"noscript","NOSCRIPT",8,i)) { // imagenes a lo largo del texto html.
               script_activo = 0;
            }
            continue;
         }
         if (script_activo) {
            continue;
         }
         if (comparar_etiquetas(texto_html,"script","SCRIPT",6,i) ||      // Inicio de script, no se tendra
             comparar_etiquetas(texto_html,"noscript","NOSCRIPT",8,i)) {  // en cuenta el texto hasta final de
            script_activo = 1;                                            // script.
            continue;
         }
         fin_etiqueta = 0;
         if (comparar_etiquetas(texto_html,"img","IMG",3,i)) {           // Se busca la etiqueta de imagen <IMG>
            while (!comparar_etiquetas(texto_html,"src","SRC",3,i)       // Buscando etiqueta <SRC="...">
                   && !fin_etiqueta) {
               if (texto_html[i] == '>') {
                  fin_etiqueta = 1;
               } else {
                  i++;
               }
            }
            if (fin_etiqueta) {
               continue;
            }
            while ((texto_html[i] != '=')) {
               i++;
            }
            i++;
            while ((texto_html[i] == ' ')) {
               i++;
            }
            if (texto_html[i] == '"') {                  // Imagen delimitada por comillas dobles
               caracter_fin_imagen = '"';
               caracter_auxiliar = '"';
               i++;
            } else if (texto_html[i] == '\'') {          // Imagen delimitada por comillas simles
               caracter_fin_imagen = '\'';
               caracter_auxiliar = '\'';
               i++;
            } else {                                     // Imagen sin delimitar, leer hasta espacio blanco
               caracter_fin_imagen = ' ';                // o hasta fin de etiqueta <IMG>
               caracter_auxiliar = '>';
            }
            contador = 0;
            imagen = (unsigned char *)malloc(TAM_NOMBRE_IMAGEN * sizeof(unsigned char));
            if (!comparar_etiquetas(texto_html,"http://","HTTP://",7,i)) {  // Si no es referencia a otro servidor
               if (texto_html[i] != '/') {
                  imagen[0] = '/';                                          // Si no empieza por '/' se le añade
                  contador = 1;
               }
            }
            while ((texto_html[i] != caracter_fin_imagen) && (texto_html[i] != caracter_auxiliar)
                   && (contador < TAM_NOMBRE_IMAGEN)) {
               imagen[contador] = texto_html[i];                           // Obtencion del path de la imagen
               contador++;
               i++;
            }
            valor_dispersion = dispersion_imagen(imagen);
            imagen_repetida = determinar_imagen_existente(tabla_hash,valor_dispersion,imagen);
            if (!imagen_repetida) {
               encolar(cola,imagen);
            }
         }
      }
      i++;
   }
   return cola;
}

/* ---------------------------------------------------------------------------------------- */
/* CONEXION_SERVIDOR: establece conexion tcp en el puerto 80 con el servidor especificado.  */
/* ---------------------------------------------------------------------------------------- */

int conexion_servidor (char *servidor, int interfaz)
{
   int sock;
   struct sockaddr_in server;
   struct hostent *info_host;
   struct in_addr **addr_host;
   char mensaje[100];
   int respuesta;
      
   // Obtencion de la direccion IP del servidor

   info_host = gethostbyname(servidor);
   if (info_host == NULL) {
      sprintf(mensaje,"Error en la obtencion de la direccion IP del servidor %s ! \n",servidor);
      imprimir_informacion_estado(&mensaje[0],interfaz);
      return -1;
   }

   // Creacion del socket TCP
   
   sock = socket(AF_INET, SOCK_STREAM, 0);
   if (sock == -1) {
      sprintf(mensaje,"Error en la creacion del socket ! \n");
      imprimir_informacion_estado(&mensaje[0],interfaz);
      return -1;
   }
   server.sin_family = AF_INET;
   server.sin_port = htons(PUERTO_HTTP);
   addr_host = (struct in_addr **)info_host -> h_addr_list;
   server.sin_addr = **addr_host;
   bzero(&(server.sin_zero),8);
   
   // Conexion al servidor

   respuesta = connect(sock, (struct sockaddr *)&server, sizeof(struct sockaddr));
   if (respuesta == -1) {
      sprintf(mensaje,"Error en la conexion al servidor %s ! \n",info_host -> h_name);
      imprimir_informacion_estado(&mensaje[0],interfaz);
      return -1;
   }
   return sock;
}

/* ---------------------------------------------------------------------------------------------- */
/* OBTENER_HOME_PAGE: obtiene la pagina HTML de inicio realizando la peticion HTTP especificada.  */
/* ---------------------------------------------------------------------------------------------- */

int obtener_home_page (int sock, char *peticion_http, char *texto_html, char *nombre_servidor, int interfaz)
{
   int longitud, respuesta;
   int bytes_leidos;
   char buffer_html[TAM_BUFFER_HTML];
   char mensaje[100];

   longitud = strlen(peticion_http);
   respuesta = send(sock,peticion_http,longitud,0);
   if (respuesta == -1) {
      sprintf(mensaje,"Error al realizar la peticion de pagina a %s ! \n",nombre_servidor);
      imprimir_informacion_estado(&mensaje[0],interfaz);
      return 0;
   }
   respuesta = 1;
   bytes_leidos = 0;
   while ((respuesta > 0) && (bytes_leidos < TAM_FICHERO_HTML)) {
      respuesta = recv(sock,buffer_html,TAM_BUFFER_HTML,0);
      if (respuesta > 0) {

         strncat(texto_html,buffer_html,respuesta);
         bytes_leidos += respuesta;
      }
   } 
   return 1;
}

/* ---------------------------------------------------------------------------------------- */
/* OBTENER_LONGITUD_IMAGEN: analiza la respuesta HTTP devuelta por el servidor y extrae la  */
/*                          longitud de la imagen recibida, ubicando el puntero del buffer  */
/*                          al comienzo de la imagen a leer.                                */
/* ---------------------------------------------------------------------------------------- */

long obtener_longitud_imagen (char *sentencia_http, int *comienzo_imagen)
{
   int contador, i = 0;
   int encontrada = 0;
   int longitud, OK = 0;
   int salto_linea;
   char tam_imagen[10];

   if (comparar_etiquetas(sentencia_http,"http","HTTP",4,i)) {
      i += 4;
      while (sentencia_http[i] != '\n' && !OK) {
         if (comparar_etiquetas(sentencia_http,"ok","OK",2,i)) {
            OK = 1;
         }
         i++;
      }
      if (OK) {
         longitud = strlen(sentencia_http);
         while (!encontrada && (i < longitud)) {
            while ((sentencia_http[i] != 'c') && (sentencia_http[i] != 'C')) {
               i++;
            }
            i++;
            if (comparar_etiquetas(sentencia_http,"ontent","ONTENT",6,i)) {
               while (sentencia_http[i] != '-') {
                  i++;
               }
               i++;
               if ((sentencia_http[i] == 'l') || (sentencia_http[i] == 'L')) {
                  i++;
                  if (comparar_etiquetas(sentencia_http,"ength","ENGTH",5,i)) {
                     while (sentencia_http[i] != ':') {
                        i++;
                     }
                     i++;
                     while (sentencia_http[i] == ' ') {
                        i++;
                     }
                     strcpy(tam_imagen,"          ");
                     contador = 0;
                     while (isdigit(sentencia_http[i])) {
                        tam_imagen[contador] = sentencia_http[i];
                        contador++;
                        i++;
                     }
                     salto_linea = 0;
                     while ((!encontrada) && (i < longitud - 2)) {
                        if(sentencia_http[i] == '\n') {
                           salto_linea ++;
                        }
                        if(salto_linea == 2) {
                           i++;
                           *comienzo_imagen = i;
                           encontrada = 1;
                           continue;
                        }
                        if((sentencia_http[i] != '\n') && (sentencia_http[i] != '\r') &&
                          (sentencia_http[i] != ' ') && (salto_linea > 0)) {
                             salto_linea--;
                        }
                        i++;
                     }
                  }
               }
            }
         }
      }
   }
   if (encontrada) {
      return atol(tam_imagen);
   } else {
      return 0;
   }
}

/* ------------------------------------------------------------------------------------------------- */
/* OBTENER_IMAGEN: realiza la peticion HTTP especificada para obtener la imagen, recibe la imagen    */
/*                 del servidor y la almacena en disco en el directorio concretado. Retornara 0      */
/*                 en caso de producirse algun error en la transferencia o almacenado de la imagen.  */
/* ------------------------------------------------------------------------------------------------- */

int obtener_imagen (int sock, char *peticion_http, char *nombre_imagen, char *directorio)
{
   int longitud, respuesta;
   unsigned char buffer_imagen[TAM_BUFFER_IMAGEN];
   char path_imagen[TAM_NOMBRE_IMAGEN];
   int archivo_imagen;
   int comienzo_imagen;
   long bytes_restantes;
   int bytes_recibidos = 1;
   long bytes_acumulados = 0;
   long tam_fichero;
   int bytes_a_leer;
   
   strcpy(path_imagen,directorio);
   strcat(path_imagen,"/");
   strcat(path_imagen,nombre_imagen);
   archivo_imagen = open(path_imagen, O_WRONLY|O_TRUNC|O_CREAT, PERMISOS_CREACION);
   if (archivo_imagen != -1) {
      longitud = strlen(peticion_http);
      respuesta = send(sock,peticion_http,longitud,0);
      if (respuesta == -1) {
         close(archivo_imagen);
         remove(path_imagen);
         return 0;
      }
      respuesta = recv(sock,buffer_imagen,TAM_BUFFER_IMAGEN,0);
      if ((respuesta == -1) || (respuesta == 0)) {
         close(archivo_imagen);
         remove(path_imagen);
         return 0;
      } 
      tam_fichero = obtener_longitud_imagen(buffer_imagen,&comienzo_imagen); 
      if (tam_fichero > 0) {
         if ((respuesta - comienzo_imagen) >= 0) {
            write(archivo_imagen,&buffer_imagen[comienzo_imagen],respuesta - comienzo_imagen);
            bytes_acumulados = respuesta - comienzo_imagen;
         }
         else {
            close(archivo_imagen);
            remove(path_imagen);
            return 0;
         }
         bytes_restantes = tam_fichero - (respuesta - comienzo_imagen);
      }
      else {
         close(archivo_imagen);
         remove(path_imagen);
         return 0;
      }
      while ((bytes_restantes > 0) && (bytes_recibidos > 0)) {
         bytes_a_leer = tam_fichero - bytes_acumulados;
         if (bytes_a_leer > TAM_BUFFER_IMAGEN) {
            bytes_a_leer = TAM_BUFFER_IMAGEN;
         }
         bytes_recibidos = recv(sock,buffer_imagen,bytes_a_leer, 0);
         if (bytes_recibidos == -1) {
            close(archivo_imagen);
            remove(path_imagen);
            return 0;
         }
         bytes_restantes -= bytes_recibidos;
         bytes_acumulados += bytes_recibidos;
         write(archivo_imagen,buffer_imagen,bytes_recibidos);
      }
      close(archivo_imagen);
   }
   else {
      return 0;
   }
   return 1;
}

/* ------------------------------------------------------------------------------------------ */
/* CONSTRUIR_PETICION_HTTP: crea una sentencia HTTP de peticion de Home Page o de imagen      */
/*                          al servidor especificado. En el caso de peticion de imagen,       */
/*                          se realizara conexion persistente para poder leer todas las       */
/*                          imagenes con un solo establecimiento y finalizacion de conexion.  */
/* ------------------------------------------------------------------------------------------ */

void construir_peticion_http (int modo, char *sentencia, char *nombre_servidor, char *nombre_elemento)
{
   
   strcpy(sentencia,"GET ");
   strcat(sentencia,nombre_elemento);
   strcat(sentencia," HTTP/1.0\n");  
   strcat(sentencia,"Accept: */*\n");
   strcat(sentencia,"Accept-Language: es\n");
   strcat(sentencia,"Host: ");
   strcat(sentencia,nombre_servidor);
   strcat(sentencia,"\nUser-Agent: Mozilla/4.0\n");
   if (modo == PETICION_IMAGEN) {
      strcat(sentencia,"Connection: Keep-Alive\n\n");
   }
   else {
      strcat(sentencia,"\n");
   }
}

/* --------------------------------------------------------------------------------------------------- */
/* COMPROBAR_IMAGEN_ENLAZADA: analiza la imagen especificada y si referencia a otro servidor devuelve  */
/*                            el nombre del servidor a conectar para obtenerla y el path relativo      */
/*                            de la imagen. Si la imagen esta en el servidor actual devuelve 0.        */
/* --------------------------------------------------------------------------------------------------- */

int comprobar_imagen_enlazada (char *imagen, char *servidor_enlace, char *imagen_enlace)
{
   int i, indice = 0;
   int fin_servidor;
   
   if (strlen(imagen) < 7) {
      return 0;
   }
   if (comparar_etiquetas(imagen,"http://","HTTP://",7,indice)) {
      indice = 7;
      fin_servidor = strcspn(&imagen[7],"/");
      for (i = 0; i < fin_servidor; i++) {
         servidor_enlace[i] = imagen[7 + i];
      }
      for (i = 0; i <= (strlen(imagen) - (fin_servidor + 7)); i++) {
         imagen_enlace[i] = imagen[7 + fin_servidor + i];
      }
      return 1;
   }
   else {
      return 0;
   }
}

/* -------------------------------------------------------------------------------------------------- */
/* SOLICITAR_IMAGENES: recorre la cola de imagenes extraidas del texto html y para cada imagen        */
/*                     ejecuta una peticion de imagen al servidor especificado por parametro.         */
/*                     Se realiza un unico establecimiento y cierre de la conexion para la obtencion  */
/*                     de todas las imagenes, mediante conexion persistente. En caso de error, se     */
/*                     vuelve a establecer la conexion con el servidor. Para imagenes ubicadas en     */
/*                     un servidor distinto al actual, se realizara una conexion adicional en la que  */
/*                     se obtendra dicha imagen.                                                      */
/* -------------------------------------------------------------------------------------------------- */

void solicitar_imagenes (char *nombre_servidor, Cola *cola, char *directorio, int interfaz)
{
   char *imagen;
   int sock, recibida;
   char sentencia[TAM_SENTENCIA_HTTP];
   char *nombre_imagen = NULL;
   int intentos = 0;
   char servidor_enlace[TAM_NOMBRE_IMAGEN];
   char imagen_enlace[TAM_NOMBRE_IMAGEN];
   int sock_aux = 0;
   char *servidor_aux = NULL;
   char mensaje[100];
   int imagenes_descargadas = 0;
   int enlazada = 0;
   
   
   reiniciar_PDI(cola);
   imagen = obtener_elemento_PDI(cola);
   if (interfaz == MODO_TEXTUAL) {
      sprintf(mensaje,"\nImagenes encontradas en la pagina solicitada de %s:\n\n",nombre_servidor);
      imprimir_informacion_estado(&mensaje[0],interfaz);
   }
   sock = conexion_servidor(nombre_servidor,interfaz);
   while (imagen != NULL)
   {
      if (comprobar_imagen_enlazada(imagen,servidor_enlace,imagen_enlace)) {
         servidor_aux = nombre_servidor;
         sock_aux = sock;
         nombre_servidor = servidor_enlace;
         imagen = imagen_enlace;
         sock = conexion_servidor(nombre_servidor,interfaz);
         enlazada = 1;
      }
      nombre_imagen = strrchr(imagen,'/') + 1;
      construir_peticion_http(PETICION_IMAGEN,sentencia,nombre_servidor,imagen);
      recibida = obtener_imagen(sock,sentencia,nombre_imagen,directorio);
      if (recibida) {
         imagenes_descargadas++;
         sprintf(mensaje,"Obtenida y almacenada imagen --> %s \n",imagen);
         if (interfaz == MODO_TEXTUAL) {
            imprimir_informacion_estado(&mensaje[0],interfaz);
         } else {
            imprimir_informacion_estado(&imagen[0],interfaz);
         }
      } else {
         if (!enlazada) {
            close(sock);
            sock = conexion_servidor(nombre_servidor,interfaz);
         }
         intentos++;
      }
      if (enlazada) {
         close(sock);
         sock = sock_aux;
         nombre_servidor = servidor_aux;
         enlazada = 0;
      }
      if (recibida || (intentos >= INTENTOS_DESCARGA_IMAGEN)) {
         avanzar_PDI(cola);
         intentos = 0;
      }
      imagen = obtener_elemento_PDI(cola);
   }
   sprintf(mensaje,"\nSe han descargado correctamente %d imagenes. \n\n",imagenes_descargadas);
   imprimir_informacion_estado(&mensaje[0],interfaz);
   close(sock);
}

/* ------------------------------------------------------------------------------------- */
/* COMPROBAR_DIRECTORIO: verifica la existencia del directorio especificad. En caso de   */
/*                       no existir el directorio, lo crea.                              */
/* ------------------------------------------------------------------------------------- */

void comprobar_directorio (char *directorio, char *subdirectorio, int modo)
{
   DIR *dir;
   
   if (modo == SUBDIRECTORIO) {
      strcat(directorio,"/");
      if (strncmp(subdirectorio,"www.",4) == 0) {
         subdirectorio += 4;
      }
      strcat(directorio,subdirectorio);
   }
   dir = opendir(directorio);
   if (dir == NULL) {
      mkdir(directorio,PERMISOS_CREACION);
   }
}

